# -*- coding: utf-8 -*-

import re
import traceback

from IPython.core.inputtransformer import InputTransformer

tests = [
    ('m0', '(1)'),
    ('m15', '((1 << 15))'),
    ('m1_2_m31', '((1 << 1) | (1 << 2) | (1 << 31))'),
    ('m4t6', '((0x7 << 4))'),
    ('m6t4', '((0x7 << 4))'),
    ('m5p1', '((0x1f << 1))'),
    ('m2p0_5_8t9_29_31t31',
     '(0x3 | (1 << 5) | (0x3 << 8) | (1 << 29) | (1 << 31))'),

    ('cafe $ m4t7', '(((0xcafe) & (0xf << 4)) >> 4)'),
    ('beef $ m2p11', '(((0xbeef) & (0x3 << 11)) >> 11)'),
    ('dead $ m3p6_m2_3',
     '(((0xdead) & (1 << 2)) >> 2) | (((0xdead) & (1 << 3)) >> 2) | '
     '(((0xdead) & (0x7 << 6)) >> 4)'),

    ('n1010', '0b1010'),
    ('b01a', '0xb01a'),
    ('110b', '0x110b'),

    ('0xabc', '0xabc'),
    ('xabcg', 'xabcg'),
    ('0000zx', '0000zx'),
    ('x01', '0x01'),
    ('fx', '0xf'),

    ('9m', '0x%x' % (9 * 1024 * 1024)),
    ('50k', '0x%x' % (50 * 1024)),
    ('128k', '0x%x' % (128 * 1024)),
    ('20g', '0x%x' % (20 * 1024 * 1024 * 1024)),

    ('0123', '0123'),
    ('a', '0xa'),
    ('12A3', '0x12A3'),
    ('123e', '0x123e'),
    ('abcg', 'abcg'),
    ('BEEF', '0xBEEF'),
]


class XalcInputTransformer(InputTransformer):

    def bitpos(self, pos):
        pos = pos.replace('m', '')

        if 't' in pos:
            start, end = pos.split('t')
            start = int(start)
            end = int(end)
            if start > end:
                start, end = end, start
        elif 'p' in pos:
            width, start = pos.split('p')
            start = int(start)
            width = int(width)
            end = start + width - 1
        else:
            start = end = int(pos)

        val = ((1 << (end - start + 1)) - 1)
        val = ('0x{:x}' if val > 1 else '{}').format(val)

        if start > 0:
            expr = '({} << {})'.format(val, start)
        else:
            expr = val

        return expr

    def replace_bit(self, match):
        poses = [self.bitpos(pos) for pos in match.group(1).split('_')]
        return '(%s)' % ' | '.join(poses)

    def get_mask_shift(self, mask):
        b = bin(eval(mask))[2:][::-1]
        return (b.index('1'), b.rindex('1'))

    def replace_extract(self, match):
        expr = match.group(1).rstrip()
        masks = [self.bitpos(pos) for pos in match.group(2).split('_')]

        maskshifts = []
        lastend = 0
        for m in sorted(masks, key=lambda m: self.get_mask_shift(m)[0]):
            start, end = self.get_mask_shift(m)
            shift = start - lastend
            maskshifts.append((m, shift))
            lastend = end - shift + 1

        fmt = '((({expr}) & {mask}) >> {shift})'

        return ' | '.join([fmt.format(expr=expr, mask=m, shift=s)
                           for m, s in maskshifts])

    def replace_size(self, match):
        sz = int(match.group(1))
        unit = match.group(2).lower()

        if unit == 'k':
            sz *= 1024
        elif unit == 'm':
            sz *= 1024 * 1024
        elif unit == 'g':
            sz *= 1024 * 1024 * 1024

        return '0x%x' % sz

    def hexrep(self, match):
        return '0x' + match.group(1)

    def binrep(self, match):
        return '0b' + match.group(1)

    def do_subs(self, line):
        reps = [
            (r'(.*)\$ m([0-9tpm_]+)\b', self.replace_extract),
            (r'\bm([0-9tpm_]+)\b', self.replace_bit),

            (r'\b(?=[0-9]*[a-fA-F]+[0-9]*)([0-9a-fA-F]+)\b', self.hexrep),

            (r'\bn([01]+)\b', self.binrep),
            (r'\b([01]+)n\b', self.binrep),

            (r'\b([0-9]+)([kKmMgG])\b', self.replace_size),

            (r'\bx([0-9a-fA-F]+)\b', self.hexrep),
            (r'\b([0-9a-fA-F]+)x\b', self.hexrep),
        ]

        for regex, rep in reps:
            line = re.sub(regex, rep, line)

        print(line)
        return line

    def push(self, line):
        try:
            return self.do_subs(line)
        except:
            traceback.print_exc()
            raise SyntaxError("Invalid syntax")

    def reset(self):
        pass



# -*- coding: utf-8 -*-

import argparse
import sys

from IPython.terminal.embed import InteractiveShellEmbed
from traitlets.config import Config

from .transformer import tests


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('cmd', nargs='*',
                        help='run this command and exit')
    args = parser.parse_args()

    c = Config()
    c.InteractiveShell.confirm_exit = False

    if args.cmd:
        c.PromptManager.out_template = ''
        c.HistoryAccessor.enabled = False

    ipshell = InteractiveShellEmbed(config=c, banner1='')
    ipshell.extension_manager.load_extension('xalc')

    if args.cmd:
        ipshell.run_cell(' '.join(args.cmd), store_history=False)
        sys.exit()

    print('xalc examples/tests:')
    for inp, out in tests:
        print(' {:10s} => {:10s}'.format(inp, out))

    ipshell()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from .transformer import XalcInputTransformer
from .formatter import XalcFormatter


def load_ipython_extension(ip):
    transformer = XalcInputTransformer()
    ip.input_transformer_manager.logical_line_transforms.append(transformer)

    formatter = ip.display_formatter.formatters['text/plain']
    formatter.for_type(int, XalcFormatter().format_int)



# -*- coding: utf-8 -*-

import re

import bitstring
import blessed


class Cursed:

    def __init__(self):
        pass

    def __getattr__(self, s):
        return ''


class XalcFormatter(object):

    def __init__(self):
        self.cursed = Cursed()
        self.t = blessed.Terminal()

    def format_size(self, sz):
        sizes = [(1024 * 1024 * 1024, 'Gi'),
                 (1024 * 1024, 'Mi'),
                 (1024, 'Ki'),
                 (1, '')]
        parts = []
        for size, name in sizes:
            if sz and sz >= size:
                parts.append('{} {}'.format(sz / size, name))
                sz -= (sz / size) * size

        return ' + '.join(parts)

    def diff_strings(self, a, b):
        diffs = []

        if len(a) != len(b):
            return diffs

        for i in range(len(a)):
            if a[i] != b[i]:
                diffs.append(i)

        return diffs

    def format_int(self, n, p, cycle):
        decfmt = ' {t.bold}{n:}{t.normal}'
        hexfmt = ' {t.bold_blue}0x{hexval}{t.normal}'

        if n >= 0:
            hexn = n
            hexval = '{n:08x}'.format(n=hexn)
        else:
            hexn = int(bitstring.BitArray('int:32={}'.format(n)).hex, 16)
            hexval = '{n:08x}'.format(n=hexn)

        fits32bits = hexn <= 0xffffffff

        decstr = decfmt.format(t=self.cursed, n=n)
        hexstr = hexfmt.format(t=self.cursed, hexval=hexval)

        declen = len(decstr)
        hexlen = len(hexstr)
        maxlen = max(declen, hexlen) + 4

        decfmt += ' ' * (maxlen - declen) + '│'
        if fits32bits:
            markers = ''.join(['{:2d}─╮ '.format(i)
                               for i in range(28, -1, -4)])
            decfmt += '  {t.white}{markers}{t.normal}'.format(
                t=self.t, markers=markers)

        hexfmt += ' ' * (maxlen - hexlen) + '│'

        if fits32bits:
            binval = ' '.join(re.findall('.{4}', '{:032b}'.format(hexn)))
            hexfmt += '  {t.magenta}{binval}{t.normal}'.format(
                t=self.t, binval=binval)
        else:
            hexfmt += '  {t.magenta}> 32 bits{t.normal}'.format(t=self.t)

        p.text(decfmt.format(t=self.t, n=n) + '\n')
        p.text(hexfmt.format(t=self.t, hexval=hexval))

        if n and n > 1024:
            p.text('\n {t.yellow}{sz:10}{t.normal}'.format(
                t=self.t, sz=self.format_size(n)))



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='xalc',
    description='Hexadecimal calculator',
    long_description=readme(),
    license='MIT',
    url='https://github.com/rabinv/xalc',
    author='Rabin Vincent',
    author_email='rabin@rab.in',
    version='0.4.0',
    install_requires=[
        'ipython>=4.2.0',
        'bitstring>=3.1.4',
        'blessed>=1.14.1',
    ],
    tests_require=['nose'],
    test_suite='nose.collector',
    entry_points={
        'console_scripts': 'xalc = xalc.xalc:main'
    },
    packages=['xalc'],
    classifiers=[
        'Environment :: Console',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development',
    ],
)
